package org.dcache.xrootd2.core;

import org.jboss.netty.channel.ChannelHandlerContext;

import org.jboss.netty.channel.MessageEvent;
import org.jboss.netty.channel.ChannelFuture;
import org.jboss.netty.channel.ChannelFutureListener;
import org.jboss.netty.handler.timeout.IdleStateAwareChannelHandler;

import org.dcache.xrootd2.protocol.messages.*;
import static org.dcache.xrootd2.protocol.XrootdProtocol.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A SimpleChannelHandler which provides an individual handler method
 * for each xrootd request type.
 *
 * Default respons to all requests is kXR_Unsupported. Sub-classes
 * may override handler methods to implement request handling.
 */
public class XrootdRequestHandler extends IdleStateAwareChannelHandler
{
    private final static Logger _log =
        LoggerFactory.getLogger(XrootdRequestHandler.class);

    public XrootdRequestHandler()
    {
    }

    @Override
    public void messageReceived(ChannelHandlerContext ctx, MessageEvent e)
    {
        AbstractRequestMessage msg = (AbstractRequestMessage) e.getMessage();

        /* FIXME: can dispatching be done in a nicer way? */
        if (msg instanceof AuthenticationRequest) {
            doOnAuthentication(ctx, e, (AuthenticationRequest) msg);
        } else if (msg instanceof LoginRequest) {
            doOnLogin(ctx, e, (LoginRequest) msg);
        } else if (msg instanceof OpenRequest) {
            doOnOpen(ctx, e, (OpenRequest) msg);
        } else if (msg instanceof StatRequest) {
            doOnStat(ctx, e, (StatRequest) msg);
        } else if (msg instanceof StatxRequest) {
            doOnStatx(ctx, e, (StatxRequest) msg);
        } else if (msg instanceof ReadRequest) {
            doOnRead(ctx, e, (ReadRequest) msg);
        } else if (msg instanceof ReadVRequest) {
            doOnReadV(ctx, e, (ReadVRequest) msg);
        } else if (msg instanceof WriteRequest) {
            doOnWrite(ctx, e, (WriteRequest) msg);
        } else if (msg instanceof SyncRequest) {
            doOnSync(ctx, e, (SyncRequest) msg);
        } else if (msg instanceof CloseRequest) {
            doOnClose(ctx, e, (CloseRequest) msg);
        } else if (msg instanceof ProtocolRequest) {
            doOnProtocolRequest(ctx, e, (ProtocolRequest) msg);
        } else if (msg instanceof RmRequest) {
            doOnRm(ctx, e, (RmRequest) msg);
        } else if (msg instanceof RmDirRequest) {
            doOnRmDir(ctx, e, (RmDirRequest) msg);
        } else if (msg instanceof MkDirRequest) {
            doOnMkDir(ctx, e, (MkDirRequest) msg);
        } else if (msg instanceof MvRequest) {
            doOnMv(ctx, e, (MvRequest) msg);
        } else if (msg instanceof DirListRequest) {
            doOnDirList(ctx, e, (DirListRequest) msg);
        } else if (msg instanceof PrepareRequest) {
            doOnPrepare(ctx, e, (PrepareRequest) msg);
        } else {
            unsupported(ctx, e, msg);
        }
    }

    protected ChannelFuture respond(ChannelHandlerContext ctx,
                                    MessageEvent e,
                                    AbstractResponseMessage msg)
    {
        return e.getChannel().write(msg);
    }

    protected ChannelFuture respondWithError(ChannelHandlerContext ctx,
                                             MessageEvent e,
                                             AbstractRequestMessage msg,
                                             int errorCode, String errMsg)
    {
        return respond(ctx, e,
                       new ErrorResponse(msg.getStreamID(), errorCode, errMsg));
    }

    protected ChannelFuture closeWithError(ChannelHandlerContext ctx,
                                           MessageEvent e,
                                           AbstractRequestMessage msg,
                                           int errorCode, String errMsg)
    {
        ChannelFuture f = respondWithError(ctx, e, msg, errorCode, errMsg);
        f.addListener(ChannelFutureListener.CLOSE);
        return f;
    }

    protected ChannelFuture unsupported(ChannelHandlerContext ctx,
                                        MessageEvent e,
                                        AbstractRequestMessage msg)
    {
        _log.info("Unsupported request: " + msg);
        return respondWithError(ctx, e, msg, kXR_Unsupported,
                                "Request " + msg.getRequestID() + " not supported");
    }

    protected void doOnLogin(ChannelHandlerContext ctx, MessageEvent e, LoginRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnAuthentication(ChannelHandlerContext ctx, MessageEvent e, AuthenticationRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnOpen(ChannelHandlerContext ctx, MessageEvent e, OpenRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnStat(ChannelHandlerContext ctx, MessageEvent e, StatRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnStatx(ChannelHandlerContext ctx, MessageEvent e, StatxRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnRead(ChannelHandlerContext ctx, MessageEvent e, ReadRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnReadV(ChannelHandlerContext ctx, MessageEvent e, ReadVRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnWrite(ChannelHandlerContext ctx, MessageEvent e, WriteRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnSync(ChannelHandlerContext ctx, MessageEvent e, SyncRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnClose(ChannelHandlerContext ctx, MessageEvent e, CloseRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnProtocolRequest(ChannelHandlerContext ctx, MessageEvent e, ProtocolRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnRm(ChannelHandlerContext ctx, MessageEvent e, RmRequest msg) {
        unsupported(ctx, e, msg);
    }

    protected void doOnRmDir(ChannelHandlerContext ctx, MessageEvent e, RmDirRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnMkDir(ChannelHandlerContext ctx, MessageEvent e, MkDirRequest msg)
    {
        unsupported(ctx, e, msg);
    }

    protected void doOnMv(ChannelHandlerContext ctx, MessageEvent e,
                            MvRequest msg) {
        unsupported(ctx, e, msg);
    }

    protected void doOnDirList(ChannelHandlerContext ctx, MessageEvent e,
                               DirListRequest msg) {
        unsupported(ctx, e, msg);
    }

    protected void doOnPrepare(ChannelHandlerContext ctx, MessageEvent e,
                               PrepareRequest msg) {
        unsupported(ctx, e, msg);
    }
}